<?php

/**
 * The public-facing functionality of the plugin.
 *
 * @link       https://lalokalabs.co
 * @since      1.0.0
 *
 * @package    Getotp
 * @subpackage Getotp/public
 */

/**
 * The public-facing functionality of the plugin.
 *
 * Defines the plugin name, version, and two examples hooks for how to
 * enqueue the public-facing stylesheet and JavaScript.
 *
 * @package    Getotp
 * @subpackage Getotp/public
 * @author     LaLoka Labs <engineering@lalokalabs.co>
 */
class Getotp_Public
{

    /**
     * The ID of this plugin.
     *
     * @since    1.0.0
     * @access   private
     * @var      string    $plugin_name    The ID of this plugin.
     */
    private $plugin_name;

    /**
     * The version of this plugin.
     *
     * @since    1.0.0
     * @access   private
     * @var      string    $version    The current version of this plugin.
     */
    private $version;

    private $getotp_options;
    private $getotp_wp_login_otp_mode;
    private $getotp_wp_login_otp_channel;
    private $getotp_woo_login_otp_mode;
    private $getotp_woo_login_otp_channel;
    private $getotp_wp_otp_phone_meta;
    private $getotp_woo_otp_phone_meta;
    private $getotp_obfuscate_mode;
    private $getotp_language;
    private $action_group;
    private $api_base_url;

    /**
     * Initialize the class and set its properties.
     *
     * @since    1.0.0
     * @param      string    $plugin_name       The name of the plugin.
     * @param      string    $version    The version of this plugin.
     */
    public function __construct($plugin_name, $version)
    {
        $this->plugin_name = $plugin_name;
        $this->version = $version;

        /*
        * Retrieve this value with:
        * $this->getotp_options = get_option( 'getotp_option_name' ); // Array of All Options
        * $getotp_api_key = $this->getotp_options['getotp_api_key']; // API KEY
        * $getotp_api_token = $this->getotp_options['getotp_api_token']; // API TOKEN
        * $getotp_wp_login_otp_mode = $this->getotp_options['getotp_wp_login_otp_mode']; // WordPress Login OTP
        * $getotp_wp_login_otp_channel = $this->getotp_options['getotp_wp_login_otp_channel']; // OTP Channel
        * $getotp_woo_login_otp_mode = $this->getotp_options['getotp_woo_login_otp_mode']; // WordPress Login OTP
        * $getotp_woo_login_otp_channel = $this->getotp_options['getotp_woo_login_otp_channel']; // OTP Channel
        */

        $this->getotp_options = get_option('getotp_option_name');

        // set getotp_wp_login_otp_mode and getotp_wp_login_otp_channel

        $getotp_wp_login_otp_mode = 'disable';
        $getotp_wp_login_otp_channel = 'email';

        if (isset($this->getotp_options['getotp_wp_login_otp_mode']) && !empty($this->getotp_options['getotp_wp_login_otp_mode'])) {
            $getotp_wp_login_otp_mode = $this->getotp_options['getotp_wp_login_otp_mode'];
        }

        if (isset($this->getotp_options['getotp_wp_login_otp_channel']) && !empty($this->getotp_options['getotp_wp_login_otp_channel'])) {
            $getotp_wp_login_otp_channel = $this->getotp_options['getotp_wp_login_otp_channel'];
        }

        $this->getotp_wp_login_otp_mode = $getotp_wp_login_otp_mode;
        $this->getotp_wp_login_otp_channel = $getotp_wp_login_otp_channel;

        // set getotp_woo_login_otp_mode and getotp_woo_login_otp_channel

        $getotp_woo_login_otp_mode = 'disable';
        $getotp_woo_login_otp_channel = 'email';

        if (isset($this->getotp_options['getotp_woo_login_otp_mode']) && !empty($this->getotp_options['getotp_woo_login_otp_mode'])) {
            $getotp_woo_login_otp_mode = $this->getotp_options['getotp_woo_login_otp_mode'];
        }

        if (isset($this->getotp_options['getotp_woo_login_otp_channel']) && !empty($this->getotp_options['getotp_woo_login_otp_channel'])) {
            $getotp_woo_login_otp_channel = $this->getotp_options['getotp_woo_login_otp_channel'];
        }

        $this->getotp_woo_login_otp_mode = $getotp_woo_login_otp_mode;
        $this->getotp_woo_login_otp_channel = $getotp_woo_login_otp_channel;

        // set getotp_wp_otp_phone_meta and getotp_woo_otp_phone_meta

        $getotp_wp_otp_phone_meta = 'mobile_phone';
        $getotp_woo_otp_phone_meta = 'billing_phone';

        if (isset($this->getotp_options['getotp_wp_otp_phone_meta']) && !empty($this->getotp_options['getotp_wp_otp_phone_meta'])) {
            $getotp_wp_otp_phone_meta = $this->getotp_options['getotp_wp_otp_phone_meta'];
        }

        if (isset($this->getotp_options['getotp_woo_otp_phone_meta']) && !empty($this->getotp_options['getotp_woo_otp_phone_meta'])) {
            $getotp_woo_otp_phone_meta = $this->getotp_options['getotp_woo_otp_phone_meta'];
        }

        $this->getotp_wp_otp_phone_meta = $getotp_wp_otp_phone_meta;
        $this->getotp_woo_otp_phone_meta = $getotp_woo_otp_phone_meta;

        // set getotp general setting

        $getotp_display_mode = 'redirect';

        if (isset($this->getotp_options['getotp_display_mode']) && !empty($this->getotp_options['getotp_display_mode'])) {
            $getotp_display_mode = $this->getotp_options['getotp_display_mode'];
        }

        $this->getotp_display_mode = $getotp_display_mode;

        // set obfuscate mode

        $getotp_obfuscate_mode = 'disable';

        if (isset($this->getotp_options['getotp_obfuscate_mode']) && !empty($this->getotp_options['getotp_obfuscate_mode'])) {
            $getotp_obfuscate_mode = $this->getotp_options['getotp_obfuscate_mode'];
        }

        $this->getotp_obfuscate_mode = $getotp_obfuscate_mode;

        // set getotp lang

        $getotp_language = 'en';

        if (isset($this->getotp_options['getotp_language']) && !empty($this->getotp_options['getotp_language'])) {
            $getotp_language = $this->getotp_options['getotp_language'];
        }

        $this->getotp_language = $getotp_language;

        $this->action_group = 'wp';

        $this->api_base_url = 'https://otp.dev';
    }

    /**
     * Register the stylesheets for the public-facing side of the site.
     *
     * @since    1.0.0
     */
    public function enqueue_styles()
    {

        /**
         * This function is provided for demonstration purposes only.
         *
         * An instance of this class should be passed to the run() function
         * defined in Getotp_Loader as all of the hooks are defined
         * in that particular class.
         *
         * The Getotp_Loader will then create the relationship
         * between the defined hooks and the functions defined in this
         * class.
         */

        wp_enqueue_style($this->plugin_name, plugin_dir_url(__FILE__) . 'css/getotp-public.css', array(), $this->version, 'all');
    }

    /**
     * Register the JavaScript for the public-facing side of the site.
     *
     * @since    1.0.0
     */
    public function enqueue_scripts()
    {

        /**
         * This function is provided for demonstration purposes only.
         *
         * An instance of this class should be passed to the run() function
         * defined in Getotp_Loader as all of the hooks are defined
         * in that particular class.
         *
         * The Getotp_Loader will then create the relationship
         * between the defined hooks and the functions defined in this
         * class.
         */

        wp_enqueue_script($this->plugin_name, plugin_dir_url(__FILE__) . 'js/getotp-public.js', array( 'jquery' ), $this->version, false);

        if ($this->getotp_woo_login_otp_mode == '2fa' && function_exists('is_woocommerce')) {
            wp_enqueue_script($this->plugin_name . '-woo-2fa-login', plugin_dir_url(__FILE__) . 'js/getotp-woo-2fa-login.js', array( 'jquery' ), $this->version, false);
        }
    }

    public function otp_login_scripts()
    {
        // if GETOTP HTTP failed, temporary disable OTP form customization

        if ($this->is_getotp_http_failed_redirect()) {
            return;
        }

        // skip if interim login
        // TODO: support interim login in future update
        if ($this->is_interim_login()) {
            return;
        }
        
        if ($this->getotp_wp_login_otp_mode == 'passwordless' && $this->recaptcha_enabled_login()) {
            wp_enqueue_script($this->plugin_name, plugin_dir_url(__FILE__) . 'js/getotp-passwordless-login.js', array( 'jquery' ), $this->version, false);
        }

        if ($this->getotp_wp_login_otp_mode == '2fa') {
            if ($this->getotp_display_mode === 'redirect') {
                wp_enqueue_script($this->plugin_name, plugin_dir_url(__FILE__) . 'js/getotp-2fa-login.js?display=redirect', array( 'jquery' ), $this->version, false);
            } elseif ($this->getotp_display_mode === 'embed') {
                wp_enqueue_script('getotp-sdk-js', $this->api_base_url . '/sdk/v1/getotp.min.js', array('jquery'), '1.0.0', false);

                // register ajax_url

                wp_enqueue_script($this->plugin_name . '-2fa-login', plugin_dir_url(__FILE__) . 'js/getotp-2fa-login.js?display=embed', array( 'jquery' ), $this->version, false);

                $this->define_wp_ajax($this->plugin_name . '-2fa-login');
            }
        }

        if ($this->getotp_wp_login_otp_channel == 'sms') {
            wp_enqueue_script($this->plugin_name, plugin_dir_url(__FILE__) . 'js/getotp-smsotp.js', array( 'jquery' ), $this->version, false);
        }
    }

    private function define_wp_ajax($script_name)
    {
        wp_add_inline_script($script_name, 'const wp_ajax = ' . json_encode(array(
            'ajax_url' => admin_url('admin-ajax.php'),
            '_nonce' => wp_create_nonce('getotp_ajax_nonce'),
        )), 'before');
    }

    /**
     * WooCommerce related. Not used until future version.
     *  */

    public function woo_otp_login_scripts()
    {
        wp_enqueue_script($this->plugin_name, plugin_dir_url(__FILE__) . 'js/getotp-woo-passwordless-login.js', array( 'jquery' ), $this->version, false);
    }

    public function wp_otp_login_form_message()
    {
        if (isset($_REQUEST['getotp_status']) && !empty($_REQUEST['getotp_status'])) {
            if ($_REQUEST['getotp_status'] === 'failed' || $_REQUEST['getotp_status'] === 'http_failed') {
                return '<div id="login_error"><strong>Error</strong>: Failed to send OTP. Please relogin.<br></div>';
            }
        }

        return;
    }

    /**
     * Customize login form
     *  */

    public function wp_otp_login_form()
    {
        // if GETOTP HTTP failed, temporary disable OTP form customization and hidden fields to bypass OTP verification

        if ($this->is_getotp_http_failed_redirect()) {
            ?>
                <input type="hidden" name="getotp_status" value="<?php echo esc_attr($_REQUEST['getotp_status']); ?>" />
                <input type="hidden" name="_getotp_nonce" value="<?php echo esc_attr($_REQUEST['_getotp_nonce']); ?>" />
            <?php

            return;
        }

        if ($this->getotp_wp_login_otp_mode === 'passwordless') {
            if (!$this->recaptcha_enabled_login()) {
                return;
            } ?>
		<style>
			#user_pass, .forgetmenot, label[for="user_pass"], .login .button.wp-hide-pw .dashicons {
                display: none;
            }
		</style>

		<?php
        }
    }

    // In order to prevent brute force attack with passwordless-login, user must enable Advanced noCaptcha & invisible plugin

    private function recaptcha_enabled_login()
    {
        return $this->is_recaptcha_enable('login');
    }
    
    private function is_recaptcha_enable($form)
    {
        if (!$form) {
            return false;
        }
        
        if (!class_exists('anr_captcha_class')) {
            return false;
        }

        $plugin_dir = WP_PLUGIN_DIR . '/advanced-nocaptcha-recaptcha';

        include_once($plugin_dir . '/functions.php');

        $form_enabled = anr_is_form_enabled($form);

        if (!$form_enabled) {
            return false;
        }
        
        $advanced_no_captcha = new anr_captcha_class();
        $show_captcha = $advanced_no_captcha->show_login_captcha();

        if (!$show_captcha) {
            return false;
        }

        return true;
    }

    /**
     * Request OTP
     *  */

    public function verify_login_with_otp($user, $username, $password)
    {
        // only run this when user submit login form (username is provided)
        
        if (empty($username)) {
            return;
        }

        if (empty($this->getotp_wp_login_otp_mode)) {
            return;
        }

        if ($this->getotp_wp_login_otp_mode === 'disable') {
            return;
        }

        // skip if interim login
        // TODO: support interim login in future update
        if ($this->is_interim_login()) {
            return;
        }

        $otp_action = 'wp_login';

        $custom_redirect = null;

        if (isset($_POST['redirect']) && ! empty($_POST['redirect'])) {
            $custom_redirect = $_POST['redirect'];
        }
        
        if ($this->is_woo_login()) {
            
            // skip woo if disable
            if ($this->is_disable_woo_login()) {
                return;
            }

            $otp_action = 'woo_login';
            $this->action_group = 'woo';
        }

        // if GETOTP HTTP failed, fallback to default login handler
        if ($this->is_getotp_http_failed_redirect()) {
            return;
        }

        $errors = new WP_Error();

        if ($this->recaptcha_enabled_login()) {
            // check valid recaptcha

            $advanced_no_captcha = new anr_captcha_class();
            $show_captcha = $advanced_no_captcha->show_login_captcha();

            $verify = $advanced_no_captcha->verify();

            if (!$show_captcha) {
                return;
            }

            if (!$verify) {
                return;
            }
        } else {
            // passwordless must recaptcha_enabled_login()
            if ($this->getotp_wp_login_otp_mode == 'passwordless') {
                return;
            }
        }

        // Check if user name or user email is valid or not
        $user = (is_email($username)) ? get_user_by('email', $username) :  get_user_by('login', $username);
        $user_id = isset($user->ID) ? $user->ID : '';

        if (empty($user_id)) {
            $errors->add('user_not_found', __('ERROR: User not found.', 'getotp'));
            return;
        }

        // validate supply username and password is valid, if not valid dont call OTP
        if ($this->getotp_wp_login_otp_mode == '2fa') {
            // get real username because username from form might be email
            $real_username = $user->user_login;

            $user_check = wp_authenticate_username_password(null, $real_username, $password);

            if (is_wp_error($user_check)) {
                $errors->add('user_not_found', __('ERROR: User not found.', 'getotp'));
                return;
            }
        }

        $lang = $this->get_language_setting();
        
        $hide_option = $this->get_obfuscate_setting();

        $otp_channel = $this->get_otp_channel($otp_action);

        $channel_value = $this->get_channel_value($otp_channel, $user);

        $final_otp_config = $this->get_final_otp_channel_and_channel_value($otp_channel, $channel_value, $user);

        $final_otp_channel = $final_otp_config['final_otp_channel'];
        $final_channel_value = $final_otp_config['final_channel_value'];

        // var_dump('$otp_channel', $otp_channel);
        // var_dump('$channel_value', $channel_value);
        // var_dump('$final_otp_channel', $final_otp_channel);
        // var_dump('$final_channel_value', $final_channel_value);
        // var_dump('$final_otp_config', $final_otp_config);
        // exit;

        $this->request_otp($user, $final_otp_channel, $final_channel_value, $hide_option, $lang, $otp_action, $custom_redirect);
    }

    /*
    handle edge case when otp_channel is sms, but the channel_value doesnt return valid mobile number
    fallback will be default channel email, with user email as value
    */
    
    private function get_final_otp_channel_and_channel_value($otp_channel, $channel_value, $user)
    {
        $phone_channels = ['sms', 'voice'];
        
        if (in_array($otp_channel, $phone_channels)) {
            if (!$this->validate_mobile_number($channel_value)) {
                $otp_channel = 'email';
                // var_dump('SMS channel with invalid mobile number ' . $channel_value . ', fallback to '. $otp_channel);

                $channel_value = isset($user->user_email) ? $user->user_email : '';
            } else {
                // format mobile number with country code and +

                $channel_value = $this->format_mobile_number($channel_value);
            }
        }
        
        return [
            'final_otp_channel' => $otp_channel,
            'final_channel_value' => $channel_value,
        ];
    }

    private function get_obfuscate_setting()
    {
        if ($this->getotp_obfuscate_mode == 'disable') {
            return 'false';
        }

        return 'true';
    }

    private function get_otp_channel($otp_action)
    {
        $otp_channel = 'email';

        $valid_channels = [
            'email',
            'sms',
            'voice',
        ];

        $channel_var = 'getotp_' .$otp_action . '_otp_channel';

        $channel_setting = $this->{$channel_var};

        if (!empty($channel_setting) && in_array($channel_setting, $valid_channels)) {
            $otp_channel = $channel_setting;
        }
        
        if ($otp_channel === 'sms') {

            // if sms channel was choose but GETOTP plan doesnt have the capability, fallback to email channel

            if (!$this->api_can_sms()) {
                $otp_channel = 'email';
                return $otp_channel;
            }
        }

        if ($otp_channel === 'voice') {

            // if sms channel was choose but GETOTP plan doesnt have the capability, fallback to email channel

            if (!$this->api_can_voice()) {
                $otp_channel = 'email';
                return $otp_channel;
            }
        }

        return $otp_channel;
    }

    private function get_channel_value($otp_channel, $user)
    {
        // default channel email
        $channel_value = isset($user->user_email) ? $user->user_email : '';

        $phone_channels = ['sms', 'voice'];

        if (in_array($otp_channel, $phone_channels)) {
            if ($this->action_group === 'wp') {
                $channel_value = get_user_meta($user->ID, $this->getotp_wp_otp_phone_meta, true);
            }

            if ($this->action_group === 'woo') {
                $channel_value = get_user_meta($user->ID, $this->getotp_woo_otp_phone_meta, true);
            }
        }
        
        return $channel_value;
    }

    private function get_instance_prefix($otp_action)
    {
        $action_exp = explode('_', $otp_action);
        $prefix = $action_exp[0];

        return $prefix;
    }

    private function get_site_language_code()
    {
        $lang = get_locale();

        $lang_exp = explode('_', $lang);

        return $lang_exp[0];
    }

    private function get_language_setting()
    {
        return $this->getotp_language;
    }

    private function api_can_sms()
    {
        // TODO: check capability from API
        return true;
    }

    private function api_can_voice()
    {
        // TODO: check capability from API
        return true;
    }

    private function format_mobile_number($phone_number)
    {
        $phone_number = ltrim($phone_number, '+');
        
        if ($this->action_group === 'wp') {
            $phone_number = '+' . $phone_number;
        }

        if ($this->action_group === 'woo') {
            // TODO: properly format using country code for woocommerce
        }
        
        return $phone_number;
    }

    private function validate_mobile_number($phone_number)
    {
        if (empty($phone_number)) {
            return false;
        }

        $phone_number = ltrim($phone_number, '+');

        if ($this->action_group === 'wp') {
            // eg: 60123456789 or +60123456789 will pass
            if (! (preg_match('/^[0-9]{11,15}$/D', $phone_number))) {
                return false;
            }
        }

        if ($this->action_group === 'woo') {
            // TODO: woo need to cater country code
            // eg: 0123456789 or +60123456789 will pass
            if (! (preg_match('/^[0-9]{9,10}$/D', $phone_number))) {
                return false;
            }
        }

        return true;
    }

    // check if authenticate filter was fired from woocommerce login
    private function is_woo_login()
    {
        if (isset($_REQUEST['woocommerce-login-nonce']) && !empty($_REQUEST['woocommerce-login-nonce'])) {
            $nonce_value = sanitize_key($_REQUEST['woocommerce-login-nonce']);

            $woo_login_nonce = wp_verify_nonce($nonce_value, 'woocommerce-login');

            // $woo_login_nonce: 1 within 12 hours, 2 within 24 hours, and false if not woocommerce login

            if ($woo_login_nonce === false) {
                return false;
            }

            return true;
        }
        
        return false;
    }

    private function is_disable_woo_login()
    {
        if (empty($this->getotp_woo_login_otp_mode) || $this->getotp_woo_login_otp_mode === 'disable') {
            return true;
        }

        return false;
    }

    // check if GETOTP API failed
    private function is_getotp_http_failed_redirect()
    {
        /*
        getotp_status values:
        1. http_failed -> error on API call
        2. failed -> because of invalid key, subscription not active or quota finished
        */
        
        if (isset($_REQUEST['getotp_status']) && $_REQUEST['getotp_status'] === 'http_failed') {
            if (isset($_REQUEST['_getotp_nonce']) && !empty($_REQUEST['_getotp_nonce'])) {
                $nonce_value = sanitize_key($_REQUEST['_getotp_nonce']);
                $getotp_nonce = wp_verify_nonce($nonce_value, 'getotp_nonce');

                if ($getotp_nonce === false) {
                    return false;
                }
        
                return true;
            }
        }

        return false;
    }

    public function ajax_login_with_otp()
    {
        $username = sanitize_text_field($_POST["username"]);
        $password = sanitize_text_field($_POST["password"]);

        
        $this->validate_user_login($username, $password);
    }

    private function validate_user_login($username, $password)
    {
        // Check if user name or user email is valid or not
        $user = (is_email($username)) ? get_user_by('email', $username) :  get_user_by('login', $username);
        $user_id = isset($user->ID) ? $user->ID : '';

        if (empty($user_id)) {
            $error = new WP_Error('001', 'No valid user information was found.');

            wp_send_json_error($error, 400);
            return;
        }
        
        $real_username = $user->user_login;

        $user_check = wp_authenticate_username_password(null, $real_username, $password);

        if (is_wp_error($user_check)) {
            $error = new WP_Error('002', 'No valid user information was found.');

            wp_send_json_error($error, 400);
            return;
        }

        $otp_action = 'wp_login';

        // parameters for request_otp

        $embed_mode = true;

        $otp_result = $this->request_otp_with_config($user, $otp_action, $embed_mode);

        die(json_encode($otp_result));
    }

    private function request_otp_with_config($user, $otp_action, $embed_mode = false)
    {
        // parameters for request_otp

        $lang = $this->get_language_setting();
        
        $hide_option = $this->get_obfuscate_setting();

        $otp_channel = $this->get_otp_channel($otp_action);

        $channel_value = $this->get_channel_value($otp_channel, $user);

        $final_otp_config = $this->get_final_otp_channel_and_channel_value($otp_channel, $channel_value, $user);

        $final_otp_channel = $final_otp_config['final_otp_channel'];
        $final_channel_value = $final_otp_config['final_channel_value'];
        
        $otp_result = $this->request_otp($user, $final_otp_channel, $final_channel_value, $hide_option, $lang, $otp_action, null, $embed_mode);

        return $otp_result;
    }

    /**
     * Authenticate with GETOTP
     *  */

    private function request_otp($user, $channel, $channel_value, $hide_option, $lang, $otp_action, $custom_redirect = null, $embed_mode = false)
    {
        $site_url = get_bloginfo('url');
        $success_url = $site_url . '?otp_status=otp_success&otp_action=' . $otp_action;

        // TODO: enable custom redirect in the future
        /* if (!empty($custom_redirect)) {
            $success_url = $success_url . '&custom_redirect=' . urlencode($custom_redirect);
        } */

        $fail_url = $site_url . '?otp_status=otp_failed';

        $getotp_payload = [
            'success_redirect_url' => $success_url,
            'fail_redirect_url' => $fail_url,
            'channel' => $channel,
            'hide' => $hide_option,
            'lang' => $lang,
        ];

        if ($embed_mode) {
            $getotp_payload['embed'] = 'compact';
        }

        if ($channel === 'email') {
            $getotp_payload['email'] = $channel_value;
        }

        if ($channel === 'sms') {
            $getotp_payload['phone_sms'] = $channel_value;
        }

        if ($channel === 'voice') {
            $getotp_payload['phone_voice'] = $channel_value;
        }
         
        $api_key = $this->getotp_options['getotp_api_key'];
        $api_token = $this->getotp_options['getotp_api_token'];
        
        $api_endpoint = esc_url_raw($this->api_base_url . '/api/verify/');

        $basic_auth = base64_encode("$api_key:$api_token");

        $args = array(
            'body'        => $getotp_payload,
            'timeout'     => '5',
            'redirection' => '5',
            'httpversion' => '1.0',
            'blocking'    => true,
            'headers'     => [
                'Authorization' => 'Basic ' . $basic_auth,
            ],
            'cookies'     => array(),
        );

        $response = wp_remote_post($api_endpoint, $args);

        $getotp_nonce = wp_create_nonce('getotp_nonce');

        if (is_wp_error($response)) {
            $error_message = $response->get_error_message();

            // if API cant be reach, fallback to default login form using redirect

            $wp_login_fallback_url = wp_login_url() . '?getotp_status=http_failed&_getotp_nonce=' . $getotp_nonce;

            wp_safe_redirect(esc_url_raw($wp_login_fallback_url));
            exit;
        } else {
            $response_code = wp_remote_retrieve_response_code($response);

            $body = wp_remote_retrieve_body($response);

            $response_data = json_decode($body, true);

            // var_dump('$response_data', $response_data);
            // var_dump('$response_code', $response_code);
            // exit;

            // handle error

            $wp_login_fallback_url = wp_login_url() . '?getotp_status=failed';

            if ($response_code == 403) {
                if (isset($response_data['detail'])) {
                    // handle credential error
                    if ($response_data['detail'] == 'Authentication credentials were not provided.') {
                        set_transient($this->plugin_name . '_crendential_error', 1);
                        set_transient($this->plugin_name . '_crendential_error_message', esc_html__($response_data['detail'], 'getotp'));

                        $this->disable_getotp_feature();

                        $this->reset_error_notice_dismissal();

                        // redirect to normal login form

                        wp_safe_redirect(esc_url_raw($wp_login_fallback_url));
                        exit;
                    }
                }
            }

            if ($response_code == 400) {

                // DRF validation error

                if (isset($response_data['email'])) {
                    $response_error_message_arr = $response_data['email'];

                    // handle invalid email

                    if ($response_error_message_arr[0] === 'Enter a valid email address.') {
                        // do something, but email always be valid
                    }

                    exit;
                }

                if (isset($response_data['phone_sms'])) {
                    $response_error_message_arr = $response_data['phone_sms'];

                    // handle invalid phone number error

                    if ($response_error_message_arr[0] === 'Invalid Phonenumber') {
                        // fallback to email OTP if user has invalid phone number
                        $this->request_otp($user, 'email', $user->user_email, $hide_option, $lang, $otp_action, $custom_redirect);
                    }
                    
                    exit;
                }

                if (isset($response_data['phone_voice'])) {
                    $response_error_message_arr = $response_data['phone_voice'];

                    // handle invalid phone number error

                    if ($response_error_message_arr[0] === 'Invalid Phonenumber') {
                        // fallback to email OTP if user has invalid phone number
                        $this->request_otp($user, 'email', $user->user_email, $hide_option, $lang, $otp_action, $custom_redirect);
                    }
                    
                    exit;
                }

                // end DRF validation error

                if (isset($response_data['code'])) {
                    $error_code = $response_data['code'];
                    $error_message = $response_data['message'];
                    $disable_getotp = false;

                    switch ($error_code) {

                        // handle lang error
                        // skip as lang value is predefine and this error will not occur
                        case 'INV-05':
                            break;

                            // handle API key error (Success URL doesn't match API user domain)
                        case 'INV-08':
                            $disable_getotp = true;
                            break;

                            // handle API key error (Fail URL doesn't match API user domain)
                        case 'INV-09':
                            $disable_getotp = true;
                            break;

                            // handle quota error
                        case 'SUB-01':
                            $disable_getotp = true;
                            set_transient($this->plugin_name . '_quota_error', 1);
                            set_transient($this->plugin_name . '_quota_error_message', esc_html__($error_message, 'getotp'));
                            break;

                            // handle channel quota error
                        case 'SUB-02':
                            break;

                            // handle subscription error
                        case 'SUB-03':
                            $disable_getotp = true;
                            set_transient($this->plugin_name . '_subscription_error', 1);
                            set_transient($this->plugin_name . '_subscription_error_message', esc_html__($error_message, 'getotp'));
                            break;
                        
                            // handle channel not supported (plan error)
                        case 'SUB-04':
                            // fallback to Email OTP if user cant send using SMS or Voice channel
                            $this->request_otp($user, 'email', $user->user_email, $hide_option, $lang, $otp_action, $custom_redirect);
                            break;

                            // handle invalid lang for current plan
                        case 'SUB-05':
                            // fallback to default language
                            $this->request_otp($user, $channel, $channel_value, $hide_option, 'en', $otp_action, $custom_redirect);
                            exit;
                    }

                    if ($disable_getotp) {
                        $this->disable_getotp_feature();
                        $this->reset_error_notice_dismissal();
                    }

                    // redirect to normal login form

                    wp_safe_redirect(esc_url_raw($wp_login_fallback_url));
                    exit;
                }
            }

            // end handle error

            // if no error, proceed to bring user to GETOTP form

            $otp_id = sanitize_key($response_data['otp_id']);

            $getotp_form_url = esc_url_raw($response_data['link']);

            update_user_meta($user->ID, 'user_otp_id', $otp_id);
            update_user_meta($user->ID, 'user_otp_activation', 0);

            if (wp_doing_ajax()) {
                return [
                    'otp_id' => $otp_id,
                    'otp_url' => $getotp_form_url,
                    'success_url' => $success_url,
                    'fail_url' => $fail_url,
                ];
            } else {
                wp_redirect($getotp_form_url);
                exit;
            }
        }
    }

    private function is_interim_login()
    {
        $interim_login = isset($_REQUEST['interim-login']);

        return $interim_login;
    }

    private function reset_error_notice_dismissal()
    {
        set_transient($this->plugin_name . '_error_notice_dismissed', 0);
    }

    private function disable_getotp_feature()
    {
        $getotp_options = $this->getotp_options;

        $getotp_options['getotp_wp_login_otp_mode'] = 'disable';
        
        update_option('getotp_option_name', $getotp_options);
    }

    public function authenticate_otp_user()
    {
        if (!isset($_GET["otp_id"]) || empty($_GET["otp_id"])) {
            return;
        }

        if (!isset($_GET["otp_status"]) || empty($_GET["otp_status"])) {
            return;
        }

        if (!isset($_GET["otp_action"]) || empty($_GET["otp_action"])) {
            return;
        }

        $otp_status = sanitize_key($_GET["otp_status"]);
        $otp_id = sanitize_key($_GET["otp_id"]);
        $otp_action = sanitize_key($_GET["otp_action"]);
        
        $custom_redirect = null;
        
        if (isset($_GET["custom_redirect"]) && !empty($_GET["custom_redirect"])) {
            $custom_redirect = sanitize_key($_GET["custom_redirect"]);

            $custom_redirect = urldecode($custom_redirect);
        }

        // validate otp_id is alpha numeric string

        if (!ctype_alnum($otp_id)) {
            return;
        }

        $valid_otp_actions = [
            'wp_login',
            'woo_login',
        ];

        if (!in_array($otp_action, $valid_otp_actions)) {
            return;
        }

        global $wpdb;

        $table = $wpdb->prefix . "usermeta";

        $meta_data	= $wpdb->get_row($wpdb->prepare("SELECT * FROM $table WHERE meta_value = %s AND meta_key = 'user_otp_id'", $otp_id));

        if (empty($meta_data)) {
            return;
        }

        $user = get_user_by('id', $meta_data->user_id);

        $user_id = $meta_data->user_id;

        // only allow otp_status with otp_success value to proceed
        
        if ($otp_status != 'otp_success') {

            // set user_otp_activation activated
            update_user_meta($user_id, 'user_otp_activation', 1);

            return;
        }

        $user_otp_activation_status = get_user_meta($user_id, 'user_otp_activation', true);

        // if not yet activate, perform otp action

        if ($user_otp_activation_status == 0) {
            if ($otp_action == 'wp_login') {
                $this->process_user_login($user, $user_id);

                // redirect to dashboard

                $dashboard_url = admin_url();

                wp_safe_redirect(esc_url_raw($dashboard_url));
                exit;
            }

            if ($otp_action == 'woo_login') {
                $this->process_user_login($user, $user_id);

                // redirect to my account page

                $redirect = wc_get_page_permalink('myaccount');

                if (!empty($custom_redirect)) {
                    $redirect = $custom_redirect;
                }

                wp_redirect(wp_validate_redirect(apply_filters('woocommerce_login_redirect', remove_query_arg('wc_error', $redirect), $user), wc_get_page_permalink('myaccount'))); // phpcs:ignore
                exit;
            }
        }
    }

    private function process_user_login($user)
    {
        $user_id = $user->ID;
        
        wp_set_current_user($user_id, $user->user_login);
        wp_set_auth_cookie($user_id);

        do_action('wp_login', $user->user_login, $user);

        // set user_otp_activation activated
        update_user_meta($user_id, 'user_otp_activation', 1);
    }

    /**
     * Customize WooCommerce login form
     *  */

    public function woo_otp_login_form()
    {
        if ($this->getotp_woo_login_otp_mode == 'passwordless' && $this->is_recaptcha_enable('woo_login')) {
            ?>
		<style>

			.woocommerce form .form-row [for="password"], .woocommerce-page form .form-row label[for="password"] {
				display:none;
			}

			.woocommerce .woocommerce-form-login .woocommerce-form-login__rememberme {
				display:none;
			}

			p.woocommerce-LostPassword {
				display:none;
			}

			#password, .show-password-input {
				display:none;

			}

		</style>

		<?php
        }
    }
}
